@@ -246,15 +246,11 @@ static int setparams_set(struct loopback_handle *lhandle,
 	} else {
 		if (lhandle == lhandle->loopback->play) {
 			val = bufsize + bufsize / 2;
-			if (val < (period_size * 3) / 4)
-				val = (period_size * 3) / 4;
 			if (val > (buffer_size * 3) / 4)
 				val = (buffer_size * 3) / 4;
 			val = buffer_size - val;
 		} else {
 			val = bufsize / 2;
-			if (val < period_size / 2)
-				val = period_size / 2;
 			if (val > buffer_size / 4)
 				val = buffer_size / 4;
 		}
@@ -949,6 +945,23 @@ static int xrun_sync(struct loopback *loop)
 			logit(LOG_CRIT, "%s start failed: %s\n", play->id, snd_strerror(err));
 			return err;
 		}
+	} else if (delay1 < fill) {
+		diff = (fill - delay1) / play->pitch;
+		while (diff > 0) {
+			delay1 = play->buf_size - play->buf_pos;
+			if (verbose > 6)
+				snd_output_printf(loop->output,
+					"sync: playback short, silence filling %li / buf_count=%li\n", (long)delay1, play->buf_count);
+			if (delay1 > diff)
+				delay1 = diff;
+			if ((err = snd_pcm_format_set_silence(play->format, play->buf + play->buf_pos * play->channels, delay1)) < 0)
+				return err;
+			play->buf_pos += delay1;
+			play->buf_pos %= play->buf_size;
+			play->buf_count += delay1;
+			diff -= delay1;
+		}
+		writeit(play);
 	}
 	if (verbose > 5) {
 		snd_output_printf(loop->output, "%s: xrun sync ok\n", loop->id);
@@ -1238,8 +1251,8 @@ static int closeit(struct loopback_handle *lhandle)
 static int init_handle(struct loopback_handle *lhandle, int alloc)
 {
 	snd_pcm_uframes_t lat;
-	lhandle->frame_size = (snd_pcm_format_width(lhandle->format) / 8) *
-							   lhandle->channels;
+	lhandle->frame_size = (snd_pcm_format_physical_width(lhandle->format) 
+						/ 8) * lhandle->channels;
 	lhandle->sync_point = lhandle->rate * 15;	/* every 15 seconds */
 	lat = lhandle->loopback->latency;
 	if (lhandle->buffer_size > lat)
@@ -1358,6 +1371,23 @@ static void lhandle_start(struct loopback_handle *lhandle)
 	lhandle->total_queued = 0;
 }
 
+static void fix_format(struct loopback *loop, int force)
+{
+	snd_pcm_format_t format = loop->capt->format;
+
+	if (!force && loop->sync != SYNC_TYPE_SAMPLERATE)
+		return;
+	if (format == SND_PCM_FORMAT_S16 ||
+	    format == SND_PCM_FORMAT_S32)
+		return;
+	if (snd_pcm_format_width(format) > 16)
+		format = SND_PCM_FORMAT_S32;
+	else
+		format = SND_PCM_FORMAT_S16;
+	loop->capt->format = format;
+	loop->play->format = format;
+}
+
 int pcmjob_start(struct loopback *loop)
 {
 	snd_pcm_uframes_t count;
@@ -1383,6 +1413,7 @@ int pcmjob_start(struct loopback *loop)
 		if (err < 0)
 			goto __error;
 		loop->play->format = loop->capt->format = err;
+		fix_format(loop, 0);
 		err = get_rate(loop->capt);
 		if (err < 0)
 			goto __error;
@@ -1394,6 +1425,7 @@ int pcmjob_start(struct loopback *loop)
 	}
 	loop->reinit = 0;
 	loop->use_samplerate = 0;
+__again:
 	if (loop->latency_req) {
 		loop->latency_reqtime = frames_to_time(loop->play->rate_req,
 						       loop->latency_req);
@@ -1442,10 +1474,19 @@ int pcmjob_start(struct loopback *loop)
 			goto __error;
 		if ((err = init_handle(loop->capt, 1)) < 0)
 			goto __error;
-		if (loop->play->rate_req != loop->play->rate)
-			loop->use_samplerate = 1;
-		if (loop->capt->rate_req != loop->capt->rate)
+		if (loop->play->rate_req != loop->play->rate ||
+                    loop->capt->rate_req != loop->capt->rate) {
+                        snd_pcm_format_t format1, format2;
 			loop->use_samplerate = 1;
+                        format1 = loop->play->format;
+                        format2 = loop->capt->format;
+                        fix_format(loop, 1);
+                        if (loop->play->format != format1 ||
+                            loop->capt->format != format2) {
+                                pcmjob_stop(loop);
+                                goto __again;
+                        }
+                }
 	}
 #ifdef USE_SAMPLERATE
 	if (loop->sync == SYNC_TYPE_SAMPLERATE)
