@@ -248,7 +248,7 @@ _("Usage: %s [OPTION]... [FILE]...\n"
 			printf(" %s", s);
 	}
 	printf(_("\nSome of these may not be available on selected hardware\n"));
-	printf(_("The availabled format shortcuts are:\n"));
+	printf(_("The available format shortcuts are:\n"));
 	printf(_("-f cd (16 bit little endian, 44100, stereo)\n"));
 	printf(_("-f cdr (16 bit big endian, 44100, stereo)\n"));
 	printf(_("-f dat (16 bit little endian, 48000, stereo)\n"));
@@ -3021,6 +3021,9 @@ static void capture(char *orig_name)
 			fd = -1;
 		}
 
+		if (in_aborting)
+			break;
+
 		/* repeat the loop when format is raw without timelimit or
 		 * requested counts of data are recorded
 		 */
